import typing
from typing import List, Optional, Tuple, Dict, Union, Callable
import numpy as np

CallableType = Callable[[str], str]
TransformType = Dict[str, Union[CallableType, float]]
TransformationDictType = Dict[str, TransformType]
TransformationListType = List[Tuple[CallableType, float]]

CONST_FUNCTION = "function"
CONST_PROBABILITY = "probability"


def lower(x: str) -> str:
    return x.lower()


def upper(x: str) -> str:
    return x.upper()


def title_first(x: str) -> str:
    return x[0].upper() + x[1:]


class Augmentor:
    """
    Apply a list of transformations as text augmentation
    """

    default_transformations: TransformationDictType = {
        "lower": {
            CONST_FUNCTION: lower,
            CONST_PROBABILITY: 0.02,
        },
        "upper": {CONST_FUNCTION: upper, CONST_PROBABILITY: 0.005},
        "title_first": {CONST_FUNCTION: title_first, CONST_PROBABILITY: 0.01},
    }

    def __init__(self, transformations: Optional[TransformationDictType] = None):
        if transformations is not None:
            self.transformation_dict = typing.cast(
                TransformationDictType, transformations
            )
        else:
            self.transformation_dict = self.default_transformations
        self.transformations = self.compile()

    def compile(self):
        transform_list: TransformationListType = []
        for transform_name, transform in sorted(self.transformation_dict.items()):
            transform_list.append(
                (transform[CONST_FUNCTION], transform[CONST_PROBABILITY])
            )
        return transform_list

    def transform(self, x: str) -> str:
        for transform, probability in self.transformations:
            if np.random.rand() < probability:
                x = transform(x)
        return x


from .core import Pretrainer
import tensorflow as tf
import numpy as np
from datasets.arrow_dataset import Dataset
from typing import List, Union


class BasePretrainer(Pretrainer):
    """
    Pretrain a CCLMBase embedder with a masked-character-like model.

    This task is to take an input, apply noisy transformations
    (including the mask), and predict the original input
    """

    def __init__(
        self,
        base=None,
        task_name="pretraining",
        load_from: str = None,
        base_args={},
        **kwargs,
    ):
        self.augmentor = kwargs.pop("augmentor", None)
        self.dropout_rate = kwargs.pop("dropout_rate", 0.1)
        super().__init__(
            base=base,
            task_name=task_name,
            load_from=load_from,
            base_args=base_args,
            **kwargs,
        )

    def get_model(self):
        """
        SDO1D + LSTM
        """
        cd = self.base.preprocessor.char_dict
        n_characters = max(list(cd.values())) + 1
        inp = self.base.embedder.input
        emb_out = self.base.embedder.output
        drop = tf.keras.layers.Dropout(0.2)
        lstm = tf.keras.layers.LSTM(128, return_sequences=True)
        dense = tf.keras.layers.Dense(
            n_characters, activation="softmax", dtype="float32"
        )
        return tf.keras.Model(inp, dense(lstm(drop(emb_out))))

    # def fit(self, data, batch_size=32):
    #     """
    #     self.model.fit(gen)
    #     """

    def generator(self, data: Union[Dataset, List[str]], batch_size: int = 32):
        """
        Return a generator over the dataset that vectorizes the text and applies
        the noisy transformation to the input.
        """
        is_dataset = isinstance(data, Dataset)
        x, y = [], []
        while True:
            for n, example in enumerate(data):
                if is_dataset:
                    example = example["text"]
                if len(example) < self.preprocessor.max_example_len:
                    continue
                substr = self.get_substr(example)
                y_str = self.preprocessor.string_to_array(
                    substr, self.preprocessor.max_example_len
                )
                if self.augmentor:
                    substr = self.augmentor.transform(substr)
                x_str = self.preprocessor.string_to_array(
                    substr, self.preprocessor.max_example_len
                )
                x_str = np.where(
                    np.random.rand(*x_str.shape) < self.dropout_rate,
                    self.preprocessor.mask_token_ind,
                    x_str,
                )
                x.append(x_str)
                y.append(y_str)
                if len(x) == batch_size or n + 1 == len(data):
                    yield np.array(x), np.array(y)
                    x, y = [], []

    def evaluate_prediction(self, x: np.ndarray, pred: np.ndarray):
        """
        evaluate the predicted string compared to the input
        """
        out_str = ""
        rev = self.preprocessor.char_rev
        maxes = np.argmax(pred, axis=2)
        for n in range(len(x)):
            out_str += f"Example {n}:\n"
            out_str += "".join(rev[i] for i in x[n]) + "\n"
            out_str += "".join(rev[i] for i in maxes[n]) + "\n\n"
        return out_str


class BasePretrainerEvaluationCallback(tf.keras.callbacks.Callback):
    def __init__(self, data: List[str], pretrainer: BasePretrainer, log_every: int = 5):
        super().__init__()
        self.pretrainer = pretrainer
        self.log_every = log_every
        self.gen = pretrainer.generator(data)

    def on_epoch_end(self, epoch, logs):
        if epoch % self.log_every == 0:
            x, y = next(self.gen)
            log_str = self.pretrainer.evaluate_prediction(x, self.model(x))
            print(log_str)
        return



from typing import List
import tensorflow as tf
import numpy as np
from ..preprocessing import Preprocessor
from ..models import CCLMModelBase, TransformerBlock


class Pretrainer:
    """
    A pretrainer needs to accept a base, implement some core layers that it
    will fit (in addition to the base, optionally), and implement a top to the
    network that represents its specific task.
    """

    def __init__(
        self,
        base=None,
        task_name="pretraining",
        load_from: str = None,
        base_args={},
        **kwargs,
    ):
        if base:
            self.preprocessor = base.preprocessor
        else:
            self.preprocessor = kwargs.get("preprocessor", Preprocessor())

        if base is None:
            base = CCLMModelBase(**base_args)
        self.base = base
        if load_from:
            self.model = tf.keras.models.load_model(load_from)
        else:
            self.model = self.get_model()
        self.task_name = task_name

    def get_model(self, *args, **kwargs):
        """
        This should return a Model that accepts input with the shape of a `base` and
        produces output of the same shape.
        """

    def compile(self, *args, **kwargs):
        self.model.compile(*args, **kwargs)

    def save_model(self, path: str):
        self.model.save(path)

    def freeze(self):
        """
        Make the layers in the model not trainable. Useful when first combining pretrained models
        with randomly initialized layers
        """
        self.model.trainable = False

    def unfreeze(self):
        """
        Make the layers in the model not trainable.
        """
        self.model.trainable = True

    def get_substr(self, inp: str) -> str:
        """
        Return a substring that is an appropriate length
        """
        max_len = self.preprocessor.max_example_len
        inp_len = len(inp)
        if inp_len <= max_len:
            return inp
        start = np.random.randint(0, inp_len - max_len)
        return inp[start : start + max_len]



from typing import List, Tuple
import tensorflow as tf
import numpy as np
from ..models import CCLMModelBase, TransformerBlock
from ..preprocessing import Preprocessor
from .core import Pretrainer


class MaskedLanguagePretrainer(tf.keras.Model, Pretrainer):
    def __init__(
        self,
        *args,
        n_conv_filters: int = 128,
        downsample_factor: int = 4,
        n_strided_convs: int = 2,
        stride_len: int = 2,
        mask_id: int = 1,
        learning_rate: float = 0.001,
        train_base: bool = True,
        training_pool_mode: str = "local",
        min_mask_len: int = 3,
        num_negatives=5,
        **kwargs,
    ):
        """
        Pretrain a model by doing masked language modeling on a corpus.

        Model that is trained accepts the base input with shape (batch_size, example_len, base_embedding_dim)
        and performs convolutions on the input. The input can be downsampled using strided conv layers,
        making the Transformer layer less expensive. The kernel_size of the conv layers
        also match the stride_len for simplicity
        """
        tf.keras.Model.__init__(self)
        self.n_strided_convs = n_strided_convs
        self.downsample_factor = downsample_factor
        self.training_pool_mode = training_pool_mode
        self.min_mask_len = min_mask_len
        self.mask_id = mask_id
        # calculate stride length to achieve the right downsampling
        assert (
            stride_len ** n_strided_convs == downsample_factor
        ), "stride_len^n_strided_convs must equal downsample_factor"
        self.stride_len = int(stride_len)
        self.n_conv_filters = n_conv_filters
        self.pool = tf.keras.layers.GlobalMaxPool1D(dtype="float32")
        self.optimizer = tf.optimizers.SGD(learning_rate)
        self.num_negatives = num_negatives

        self.train_base = train_base

        Pretrainer.__init__(self, *args, **kwargs)

        # initialize output weights for the sampled softmax or nce loss
        self.output_weights = tf.Variable(
            tf.random.normal(
                [self.preprocessor.tokenizer.get_vocab_size() + 1, self.n_conv_filters]
            )
        )
        self.output_biases = tf.Variable(
            tf.zeros([self.preprocessor.tokenizer.get_vocab_size() + 1])
        )
        # negative sampling head
        self.classification_head = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(128, activation="tanh"),
                tf.keras.layers.Dense(1, activation="sigmoid", dtype="float32"),
            ]
        )
        self.concat = tf.keras.layers.Concatenate()
        self.output_embedding = tf.keras.layers.Embedding(
            self.preprocessor.tokenizer.get_vocab_size() + 1, self.n_conv_filters
        )
        self.pretraining_model = self.get_pretraining_model()

    def get_model(
        self,
    ):
        """
        Until handled better, inputs need to be padded to a multiple of filter_stride_len*n_strided_convs.

        The model uses one or more strided Conv1D to reduce the input shape before passing it to
        one or more transformer blocks
        """
        # reduce the size, transformer, upsample

        layers = [
            tf.keras.layers.Conv1D(
                self.n_conv_filters,
                self.stride_len,
                strides=self.stride_len,
                padding="same",
                activation="tanh",
            )
            for _ in range(self.n_strided_convs)
        ]
        model = tf.keras.Sequential(
            [
                tf.keras.layers.Conv1D(
                    self.n_conv_filters,
                    self.stride_len,
                    padding="same",
                    activation="tanh",
                ),
                tf.keras.layers.Dropout(0.2),
                *layers,
                tf.keras.layers.Dropout(0.2),
                TransformerBlock(embed_dim=self.n_conv_filters),
                *[
                    tf.keras.layers.UpSampling1D(size=self.downsample_factor)
                    for _ in range(1)
                    if self.downsample_factor > 1
                ],
                tf.keras.layers.Conv1D(
                    self.n_conv_filters,
                    self.stride_len,
                    padding="same",
                    activation="tanh",
                ),
            ]
        )

        return model

    def can_learn_from(self, example_str: str) -> bool:
        """
        Decide whether it's appropriate to learn from this example
        """
        # if it's an empty string, skip it
        if example_str == "":
            return False

        # if it's just a [CLS] a few tokens and [SEP], skip it
        if len(example_str) < 20:
            return False
        return True

    def batch_from_strs(
        self, input_strs: List[str]
    ) -> Tuple[List[str], List[Tuple[int, int, str]], List[int]]:
        """
        Transform input strings into correct-length substrings and pick tokens to mask
        """
        batch_inputs: List[str] = []
        batch_outputs: List[int] = []
        batch_spans: List[Tuple[int, int, str]] = []
        tokenizer = self.preprocessor.tokenizer
        for example in input_strs:

            # subset to a substring of the correct len
            # encoded = self.get_substr(encoded)
            example = self.get_substr(example)
            encoded = tokenizer.encode(example)

            # get all tokens that are long enough to be masked
            possible_masked_tokens = [
                _id
                for n, _id in enumerate(encoded.ids)
                if len(encoded.tokens[n]) >= self.min_mask_len
            ]
            possible_encoding_indexes = [
                n
                for n, token in enumerate(encoded.tokens)
                if len(token) >= self.min_mask_len
            ]
            # if none, pick a shorter token
            if len(possible_masked_tokens) == 0:
                possible_masked_tokens = encoded.ids
                possible_encoding_indexes = [n for n in range(0, len(encoded.ids))]
            # sample a value to index into possible_masked_tokens
            masked_token_index = np.random.randint(0, len(possible_masked_tokens))
            # look up the value sampled
            masked_token_id = possible_masked_tokens[masked_token_index]
            start, end = encoded.token_to_chars(
                possible_encoding_indexes[masked_token_index]
            )
            masked_token_len = end - start
            inp = (
                example[:start]
                + "?" * masked_token_len
                + example[start + masked_token_len :]
            )
            batch_inputs.append(inp)
            batch_spans.append((start, end, example[start : start + masked_token_len]))
            batch_outputs.append(masked_token_id)
        return batch_inputs, batch_spans, batch_outputs

    def generator(self, data: List[str], batch_size: int):
        """
        Generator for training purposes
        """
        while True:
            batch_inputs, batch_outputs, batch_spans, batch_strs = (
                [],
                [],
                [],
                [],
            )
            for n, example in enumerate(data):
                example = example.strip()

                if not self.can_learn_from(example):
                    continue
                batch_strs.append(example)
                if len(batch_strs) == batch_size or (
                    n + 1 == len(data) and len(batch_strs) > 0
                ):
                    batch_inputs, batch_spans, batch_outputs = self.batch_from_strs(
                        batch_strs
                    )

                    x = self.get_batch(batch_inputs, batch_spans=batch_spans)
                    # x = tf.concat([x, *[x for _ in range(self.num_negatives)]], axis=0)

                    y = np.array(batch_outputs)
                    y_sample, token_sample, mask = self.sample_from_positive(
                        y, x, batch_spans
                    )
                    yield [x, token_sample, mask], y_sample
                    batch_inputs, batch_outputs, batch_spans, batch_strs = (
                        [],
                        [],
                        [],
                        [],
                    )

    def train_step(self, data):
        """
        Iterate over a corpus of strings, using the preprocessor's tokenizer to mask
        some tokens, and predicting the masked tokens.
        """
        x, y_true = data
        x_char, x_token, mask = x
        loss, preds = self.train_batch(x_char, x_token, y_true, mask)
        self.compiled_metrics.update_state(y_true, preds)
        return {m.name: m.result() for m in self.metrics}

    def account_for_downsample(self, inp_len: int) -> int:
        """
        Since the transformer model might downsample, we might need to do extra padding
        """
        return inp_len + (inp_len % self.downsample_factor)

    def get_batch(
        self, input_strings: List[str], batch_spans: List[Tuple[int, int, str]] = None
    ) -> np.ndarray:
        """
        Return vector encoding of a batch of strings using the preprocessor
        """
        max_len = self.account_for_downsample(np.max(list(map(len, input_strings))))
        out = np.zeros((len(input_strings), max_len))
        for n, s in enumerate(input_strings):
            out[n] = self.preprocessor.string_to_array(s, max_len)
            if batch_spans is not None:
                start, end, _n = batch_spans[n]
                out[n][start:end] = self.mask_id
        return out

    def train_batch(
        self, x_char: np.ndarray, x_token: np.ndarray, y: np.ndarray, mask: np.ndarray
    ):
        with tf.GradientTape() as g:
            # char_rep = self.get_embedding_for_pretraining(
            #     x_char, mask=mask, training=True
            # )
            # token_emb = self.output_embedding(x_token)

            # loss, predictions = self.get_loss(char_rep, token_emb, y)
            predictions = self([x_char, x_token, mask], training=True)
            loss = self.compiled_loss(y, predictions)
            to_diff = (
                self.model.trainable_weights + self.base.embedder.trainable_weights
                if self.train_base
                else []
                + [
                    self.output_embedding.trainable_weights,
                    self.classification_head.trainable_weights,
                ]
            )
            gradients = g.gradient(loss, to_diff)
            self.optimizer.apply_gradients(zip(gradients, to_diff))

        return loss, predictions

    def get_embedding_for_pretraining(
        self, x: np.ndarray, mask, training=True
    ) -> np.ndarray:
        """
        Embed the text and possibly pool the whole thing or pool just the masked token
        """
        rep = self.model(self.base.embedder(x, training=training), training=training)
        # either pool the whole representation
        if self.training_pool_mode == "global":
            rep = self.pool(rep)
        else:
            rep = self.pool(self.mask_mul([rep, mask]))
        return rep

    def sample_from_positive(self, y, x, batch_spans):
        with tf.device("/GPU:0"):
            # Compute the average loss for the batch.
            # y_true = tf.cast(tf.reshape(y, [-1, 1]), tf.int64)
            y_true = y.reshape(-1, 1)
            # sampled = tf.random.learned_unigram_candidate_sampler(
            (
                neg_sample,
                true_expected,
                neg_expected,
            ) = tf.random.log_uniform_candidate_sampler(
                y_true,
                1,
                self.num_negatives * y_true.shape[0],
                True,
                self.preprocessor.tokenizer.get_vocab_size() + 1,
                name="sampler_dist",
            )
            # true samples get a 1, negative samples get a 0
            y_sample = np.zeros((y_true.shape[0], 1 + self.num_negatives))
            y_sample[:, 0] = 1.0
            # y_sample = tf.concat(
            #     (tf.ones((y_true.shape[0], 1)), tf.zeros((neg_sample.shape[0], 1))),
            #     axis=0,
            # )
            neg_sample = neg_sample.numpy().reshape(y_true.shape[0], self.num_negatives)
            token_sample = np.concatenate((y_true, neg_sample), axis=1)
            # token_sample = tf.reshape(
            #     tf.concat((y_true, tf.reshape(neg_sample, (-1, 1))), axis=0), (-1,)
            # )
            mask = np.zeros((*(x.shape), self.n_conv_filters))
            ind = 0
            for span in batch_spans:
                mask[ind][span[0] : span[1]] = 1
                ind += 1

            # for _ in range(self.num_negatives):
            #     for span in batch_spans:
            #         mask[ind][span[0] : span[1]] = 1
            #         ind += 1
            # mask = mask.reshape((*mask.shape[:2], 1))
            # mask = tf.cast(mask, tf.float16)
            return y_sample, token_sample, mask

    def get_loss(
        self,
        rep_char: tf.Tensor,
        rep_token: tf.Tensor,
        y_sample: np.ndarray,
    ):
        with tf.device("/GPU:0"):

            cat = self.concat([rep_char, rep_token])
            predictions = self.classification_head(cat)
            loss = self.compiled_loss(y_sample, predictions)
            return loss, predictions

    def get_pretraining_model(self):
        """
        Return a Model for training using negative sampling
        """
        inp_char = tf.keras.layers.Input((None,))
        inp_token = tf.keras.layers.Input((1 + self.num_negatives,))
        inp_mask = tf.keras.layers.Input((None,))

        rep = self.base.embedder(inp_char)
        rep = self.model(rep)
        mul = self.pool(tf.keras.layers.Multiply()([rep, inp_mask]))
        mul = tf.keras.layers.Reshape((1, self.n_conv_filters))(mul)

        emb_token = self.output_embedding(inp_token)
        # cat = self.concat([emb_token, mul])
        # out = self.classification_head(cat)
        out = tf.keras.layers.Activation("linear", dtype="float32")(
            tf.keras.layers.Dot((2, 2), normalize=True)([emb_token, mul])
        )
        return tf.keras.Model([inp_char, inp_token, inp_mask], out)

    def call(self, x, training=False):
        return self.pretraining_model(x, training=training)


class MLMTrainingLogger:
    def __init__(self, generator, preprocessor):
        self.generator = generator
        self.preprocessor = preprocessor
        self.logged_examples = []

    def log_example(self, x_char, x_token, y_pred):
        example_text = "".join([self.preprocessor.char_rev[i] for i in x_char])
        candidates = "\n".join(
            [self.preprocessor.tokenizer.id_to_token(i) for i in x_token]
        )
        return f"Example:\n\n{example_text}\n\nCandidates:\n{candidates}\n\nPredictions: {y_pred}"


from tqdm import tqdm
import numpy as np
from tokenizers import Encoding, Tokenizer
from tokenizers.models import BPE
from tokenizers.trainers import BpeTrainer
from tokenizers.pre_tokenizers import Whitespace
from tokenizers.normalizers import Lowercase, NFD, StripAccents, Sequence
from typing import List, Union
from datasets.arrow_dataset import Dataset
import json
import os
import re
from collections import Counter
from typing import Dict, List, Iterable

PUNCT = "~!@#$%^&*()-+=_.,?\":;'"


class Preprocessor:
    def __init__(
        self,
        load_from: str = None,
        vocab_size: int = 10000,
        max_example_len: int = 128,
        batch_size: int = 16,
        num_stopwords: int = 250,
        mask_output_len: int = 4,
        mask_token_ind: int = 1,
    ):
        self.char_dict: Dict[str, int] = {}
        self.char_rev: Dict[int, str] = {}
        self.token_dict: Dict[str, int] = {}
        self.token_rev: Dict[str, int] = {}
        self.vocab_size = vocab_size
        self.max_example_len = max_example_len
        self.batch_size = batch_size
        self.num_stopwords = num_stopwords
        self.mask_output_len = mask_output_len
        self.mask_token_ind = mask_token_ind
        self.tokenizer_is_fit = False
        self.tokenizer = Tokenizer(BPE(unk_token="[UNK]"))
        self.tokenizer.pre_tokenizer = Whitespace()
        self.tokenizer.normalizer = Sequence([NFD(), Lowercase(), StripAccents()])
        self.tok_trainer = BpeTrainer(
            special_tokens=["[UNK]", "[MASK]"], vocab_size=self.vocab_size
        )
        if load_from:
            self._load(load_from)

    def fit(
        self,
        data: Union[List[str], Dataset],
        min_char_count: int = 1,
        skip_tokenizer: bool = False,
        progbar: bool = True,
    ):
        """
        Create a character-level dictionary based on a list of strings and train
        the tokenizer on the dataset. Training the tokenizer allows for using it
        in the language modeling objective
        """
        is_dataset = isinstance(data, Dataset)
        if not self.tokenizer_is_fit and not skip_tokenizer:
            print("fitting tokenizer")
            self.tokenizer.train_from_iterator(data, trainer=self.tok_trainer)
            self.tokenizer_is_fit = True
        char_counter: Counter = Counter()
        token_counter: Counter = Counter()
        iterator_: Iterable = data
        if progbar:
            iterator_ = tqdm(data)
        print("fitting char_dict")
        for example in iterator_:
            if is_dataset:
                example = example["text"]
            chars = Counter(example)
            for char, char_count in chars.items():
                try:
                    char_counter[char] += char_count
                except KeyError:
                    char_counter[char] = char_count

        counts = [k for k, v in char_counter.items() if v >= min_char_count]
        self.char_rev = {0: "", 1: "?", 2: "?", 3: ""}

        for c in sorted(counts):
            n = len(self.char_rev)
            self.char_rev[n] = c
            self.char_dict[c] = n

    def tokenize(self, string_to_tokenize: str) -> Encoding:
        string_to_tokenize = string_to_tokenize
        return self.tokenizer.encode(string_to_tokenize)

    def string_to_array(self, string_in: str, length: int, padding_pre: bool = False):
        # truncate
        if padding_pre:
            s = string_in[-length:]
        else:
            s = string_in[:length]
        # map char -> int and left-zero-pad
        mapped = np.ones((len(s)))
        for n, char in enumerate(s):
            try:
                mapped[n] = self.char_dict[char]
            except KeyError:
                pass
        # mapped = [self.char_dict.get(x, 1) for x in s]
        if padding_pre:
            r = np.pad(mapped, (length - len(s), 0), "constant", constant_values=(0, 0))
        else:
            r = np.pad(mapped, (0, length - len(s)), "constant", constant_values=(0, 0))
        return r

    def save(self, path: str):
        """
        Write a Preprocessor object to a .JSON config
        """
        config = {
            "char_rev": self.char_rev,
            "char_dict": self.char_dict,
            "max_example_len": self.max_example_len,
        }
        with open(os.path.join(path, "cclm_config.json"), "w") as f:
            json.dump(config, f)

    def _load(self, path: str):
        """
        Load a Preprocessor object from disk
        """
        with open(path, "rb") as f:
            result = json.load(f)
        for key, value in result.items():
            setattr(self, key, value)
        for key, value in list(self.char_rev.items()):
            self.char_rev[int(key)] = value



from typing import Any, Generator, List, Dict
import numpy as np
import json
import os
from .preprocessing import Preprocessor
import tensorflow as tf
from tensorflow.keras.mixed_precision import experimental as mixed_precision

policy = mixed_precision.Policy("mixed_float16")
mixed_precision.set_policy(policy)


DEFAULT_EMB_DIM = 128


class GlobalBroadcaster1D(tf.keras.layers.Layer):
    def __init__(self, dim: int):
        super().__init__()
        self.pooler = tf.keras.layers.GlobalMaxPool1D()
        self.reshaper = tf.keras.layers.Reshape((1, -1))
        self.dense = tf.keras.layers.Dense(dim, activation="relu")
        self.cat = tf.keras.layers.Concatenate()

    def call(self, inputs, training):
        # globally pool and broadcast back
        side = self.reshaper(self.pooler(inputs))
        side = tf.repeat(self.dense(side), tf.shape(inputs)[1], axis=1)
        return self.cat([inputs, side])


def zero_out_zero_embedding(model: tf.keras.Model):
    emb_layers = [l for l in model.layers if l.name.find("embedding") > -1]
    for layer in model.layers:
        if isinstance(layer, tf.keras.layers.Embedding):
            print(f"zeroing out {layer.name}")
            w = layer.get_weights()
            w[0][0] = np.zeros_like(w[0][0])
            layer.set_weights(w)


def get_block(
    x: tf.Tensor,
    n_filters: int,
    filter_len: int,
    global_filters: int,
    prefix: str,
    suffix: str,
):
    """
    Add a block of layers on top of input tensor
    """
    char_conv_1 = tf.keras.layers.Conv1D(
        n_filters,
        filter_len,
        activation=tf.keras.layers.LeakyReLU(alpha=0.1),
        name=f"{prefix}_conv_{suffix}_1",
        padding="same",
    )
    char_conv_2 = tf.keras.layers.Conv1D(
        n_filters,
        filter_len,
        activation=tf.keras.layers.LeakyReLU(alpha=0.1),
        name=f"{prefix}_conv_{suffix}_2",
        padding="same",
    )
    char_conv_3 = tf.keras.layers.Conv1D(
        n_filters,
        filter_len,
        activation=tf.keras.layers.LeakyReLU(alpha=0.1),
        name=f"{prefix}_conv_{suffix}_3",
        padding="same",
    )
    broadcaster_1 = GlobalBroadcaster1D(global_filters)
    cc1 = char_conv_1(x)
    x = tf.keras.layers.Dropout(0.25)(cc1)
    x = char_conv_2(x)
    x = broadcaster_1(x)
    x = char_conv_3(x)
    return tf.keras.layers.Add(name=f"{prefix}_res_{suffix}")([cc1, x])


def get_character_embedder(
    max_len: int,
    char_emb_size: int,
    n_blocks: int,
    n_chars: int,
    n_filters: int,
    prefix: str,
    global_filters: int = 64,
) -> tf.keras.Model:
    """
    Return a basic model that embeds character-level input and passes it through conv
    layers that don't change its length.

    Also aggregated sequence-level info by pooling then multiplying to make global information
    available in lower levels

    TODO: make this into a class so it's more obvious how to write your own
    """
    inp = tf.keras.layers.Input((max_len,), name=f"{prefix}_inp")
    char_emb = tf.keras.layers.Embedding(
        n_chars,
        char_emb_size,
        name=f"{prefix}_embedding",
    )
    # char_conv_1 = tf.keras.layers.Conv1D(
    #     filters,
    #     3,
    #     activation=tf.keras.layers.LeakyReLU(alpha=0.1),
    #     name=f"{prefix}_c1",
    #     padding="same",
    # )
    # char_conv_2 = tf.keras.layers.Conv1D(
    #     filters,
    #     3,
    #     activation=tf.keras.layers.LeakyReLU(alpha=0.1),
    #     name=f"{prefix}_c2",
    #     padding="same",
    # )
    # char_conv_3 = tf.keras.layers.Conv1D(
    #     filters,
    #     3,
    #     activation=tf.keras.layers.LeakyReLU(alpha=0.1),
    #     name=f"{prefix}_c3",
    #     padding="same",
    # )
    # char_conv_4 = tf.keras.layers.Conv1D(
    #     filters,
    #     3,
    #     activation=tf.keras.layers.LeakyReLU(alpha=0.1),
    #     name=f"{prefix}_c4",
    #     padding="same",
    # )
    # char_conv_5 = tf.keras.layers.Conv1D(
    #     4 * filters,
    #     3,
    #     activation=tf.keras.layers.LeakyReLU(alpha=0.1),
    #     name=f"{prefix}_c5",
    #     padding="same",
    # )
    # broadcaster_1 = GlobalBroadcaster1D(64)
    # broadcaster_2 = GlobalBroadcaster1D(64)

    # initial input and conv
    x = char_emb(inp)
    # cc1 = char_conv_1(x)
    # x = tf.keras.layers.Dropout(0.25)(cc1)
    # x = char_conv_2(x)
    # x = broadcaster_1(x)
    # x = char_conv_3(x)
    # r1 = tf.keras.layers.Add(name=f"{prefix}_res_conn")([cc1, x])
    # x = tf.keras.layers.Dropout(0.25)(x)
    # x = char_conv_4(x)
    # r2 = tf.keras.layers.Add(name=f"{prefix}_res_conn_2")([r1, x])
    # x = broadcaster_2(r2)
    # x = char_conv_5(x)
    for n in range(n_blocks):
        x = get_block(
            x,
            n_filters,
            3,
            global_filters,
            prefix,
            str(n),
        )
    return inp, x


class CCLMModelBase:
    def __init__(
        self,
        load_from=None,
        preprocessor=None,
        char_emb_size=32,
        n_filters=256,
        prefix="cclm",
        pool=False,
    ):
        self.char_emb_size = char_emb_size
        self.n_filters = n_filters
        self.prefix = prefix
        self.preprocessor = preprocessor
        if load_from:
            self._load(load_from, pool=pool)
        else:
            emb_in, emb_out = get_character_embedder(
                self.preprocessor.max_example_len,
                char_emb_size,
                2,
                np.max(list(self.preprocessor.char_dict.values())) + 1,
                n_filters,
                prefix,
            )
            self.embedder = tf.keras.Model(emb_in, emb_out)

    def save(self, path: str):
        """
        Use keras to save the model and its preprocessor
        """
        self.embedder.save(path)

    def freeze_embedder(self):
        self.embedder.trainable = False

    def unfreeze_embedder(self):
        self.embedder.trainable = True

    def _load(self, path: str, pool: bool = False):
        """
        Load a model and its preprocessor
        """
        emb_in, emb_out = get_character_embedder(
            self.preprocessor.max_example_len,
            self.char_emb_size,
            2,
            np.max(list(self.preprocessor.char_dict.values())),
            self.n_filters,
            self.prefix,
        )
        if pool:
            emb_out = tf.keras.layers.GlobalMaxPooling1D(name="context_gmp")(emb_out)
        self.embedder = tf.keras.Model(emb_in, emb_out)
        loaded_model = tf.keras.models.load_model(path)
        loaded_model.compile("Adam", "binary_crossentropy")
        print([layer.name for layer in loaded_model.layers])
        for layer in loaded_model.layers[:-1]:
            try:
                old_weights = layer.get_weights()
                if len(old_weights) == 0:
                    continue
                print(f"initializing layer: {layer.name}")
                new_layer = self.embedder.get_layer(layer.name)
                new_layer.set_weights(old_weights)
            except ValueError:
                print(f"unable to transfer weights for {layer.name}")


def rev(prep: Preprocessor, a: str):
    return "".join([prep.char_rev[int(c)] for c in a])


class PrinterCallback(tf.keras.callbacks.Callback):
    def __init__(self, gen: Generator, prep: Preprocessor):
        super().__init__()
        self.gen = gen
        self.prep = prep

    def on_epoch_end(self, epoch: int, logs: Dict[str, Any]):
        j = 0
        ex = next(self.gen)
        sourceex = ex[0][0]
        print("*" * 80)
        label = ex[1][j]
        posex = ex[0][j]
        print(
            rev(self.prep, posex),
            ":",
            self.model.predict(ex[0][j : j + 1]),
            " -> ",
            label,
        )
        print("*" * 80)


class MultiHeadSelfAttention(tf.keras.layers.Layer):
    """
    From keras examples
    https://keras.io/examples/nlp/text_classification_with_transformer/
    """

    def __init__(self, embed_dim, num_heads=8):
        super(MultiHeadSelfAttention, self).__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        if embed_dim % num_heads != 0:
            raise ValueError(
                f"embedding dimension = {embed_dim} should be divisible by number of heads = {num_heads}"
            )
        self.projection_dim = embed_dim // num_heads
        self.query_dense = tf.keras.layers.Dense(embed_dim)
        self.key_dense = tf.keras.layers.Dense(embed_dim)
        self.value_dense = tf.keras.layers.Dense(embed_dim)
        self.combine_heads = tf.keras.layers.Dense(embed_dim)

    def attention(self, query, key, value):
        score = tf.matmul(query, key, transpose_b=True)
        dim_key = tf.cast(tf.shape(key)[-1], tf.float16)
        scaled_score = score / tf.math.sqrt(dim_key)
        weights = tf.nn.softmax(scaled_score, axis=-1)
        output = tf.matmul(weights, value)
        return output, weights

    def separate_heads(self, x, batch_size):
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.projection_dim))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, inputs):
        # x.shape = [batch_size, seq_len, embedding_dim]
        batch_size = tf.shape(inputs)[0]
        query = self.query_dense(inputs)  # (batch_size, seq_len, embed_dim)
        key = self.key_dense(inputs)  # (batch_size, seq_len, embed_dim)
        value = self.value_dense(inputs)  # (batch_size, seq_len, embed_dim)
        query = self.separate_heads(
            query, batch_size
        )  # (batch_size, num_heads, seq_len, projection_dim)
        key = self.separate_heads(
            key, batch_size
        )  # (batch_size, num_heads, seq_len, projection_dim)
        value = self.separate_heads(
            value, batch_size
        )  # (batch_size, num_heads, seq_len, projection_dim)
        attention, weights = self.attention(query, key, value)
        attention = tf.transpose(
            attention, perm=[0, 2, 1, 3]
        )  # (batch_size, seq_len, num_heads, projection_dim)
        concat_attention = tf.reshape(
            attention, (batch_size, -1, self.embed_dim)
        )  # (batch_size, seq_len, embed_dim)
        output = self.combine_heads(
            concat_attention
        )  # (batch_size, seq_len, embed_dim)
        return output


class TransformerBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int = DEFAULT_EMB_DIM,
        num_heads: int = 8,
        ff_dim: int = 32,
        rate=0.1,
    ):
        super(TransformerBlock, self).__init__()
        self.att = MultiHeadSelfAttention(embed_dim, num_heads)
        self.ffn = tf.keras.Sequential(
            [
                tf.keras.layers.Dense(ff_dim, activation="relu"),
                tf.keras.layers.Dense(embed_dim),
            ]
        )
        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def call(self, inputs, training):
        attn_output = self.att(inputs)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        return self.layernorm2(out1 + ffn_output)


class ComposedModel(tf.keras.layers.Layer):
    def __init__(
        self,
        base: CCLMModelBase,
        models: List[tf.keras.Model],
        name: str = "composed_model",
    ):
        """
        Holds a set of pretrained models that share a base. When called, it will
        run the input through the base's embedder, call each model on that embedding,
        and concatenate the output along the last dimension
        """
        super().__init__()
        self.models = models
        self.base = base
        self.model_name = name
        self.model = self.build_model()

    def build_model(self):
        emb = self.base.embedder

        towers = []
        for model in self.models:
            x = model(emb.output)
            towers.append(x)
        cat = tf.keras.layers.Concatenate()(towers)
        return tf.keras.Model(emb.input, cat, name=self.model_name)

    def call(self, inputs, training):
        return self.model(inputs, training=training)



import os
from setuptools import setup, find_packages

setup(
    name="cclm",
    version="0.0.1",
    author="jamesmf",
    author_email="",
    description=("composable character level models"),
    license="BSD",
    keywords="embeddings composable character-level",
    url="",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
    install_requires=[
        "tokenizers",
        "datasets",
        "tensorflow",
        "numpy",
        "tqdm",
        "pytest",
    ],
)
